import yaml

DEFAULT_CONFIG_PATH = '.bibl.yml'
_config = dict()


def load_config(file):
    global _config
    if not _config:
        _load_config(DEFAULT_CONFIG_PATH)
    _load_config(file)


def get_config():
    global _config
    if not _config:
        _load_config(DEFAULT_CONFIG_PATH)
    return _config


def set_config(key, value):
    global _config
    if not value is None:
        _config[key] = value
        if key in {'select', 'ignore'} and _config[key] is None:
            _config[key] = []
        _validate_config(_config)


def _load_config(file):
    global _config
    with open(file) as config_file:
        config = yaml.load(config_file, Loader=yaml.FullLoader)
    for k, v in config.items():
        set_config(k, v)


def _validate_config(config):
    if 'select' in config and 'ignore' in config and config['select'] and config['ignore']:
        raise ValueError(
            "Configuration cannot contain both included and selected and ignored rules. Use either include or exclude to select"
            "enabled rules."
        )



import logging
import re
import sys
from dataclasses import dataclass

from pybtex.database import parse_file

from bibl.rule import load_rules, Rule
from bibl.text_utils import find_entry_line_number, MONTH_NAMES


def lint(bibliography, output=True):
    bib_data = parse_file(bibliography, macros=MONTH_NAMES)
    bib_data.file = bibliography
    with open(bibliography, 'r') as bib_file:
        bib_text = bib_file.read()

    rules = load_rules()

    warnings = []

    _apply_text_rules(bibliography, bib_text, rules.enabled_text_rules, warnings)
    _apply_entry_rules(bibliography, bib_data, bib_text, rules.enabled_entry_rules, warnings)

    warnings.sort(key=lambda w: w.rule.id)
    warnings.sort(key=lambda w: w.line)
    if output:
        for warning in warnings:
            warning.log()
    return warnings


def _apply_text_rules(bibliography, bib_text, text_rules, warnings):
    for i, line in enumerate(bib_text.split('\n')):
        line_number = i + 1
        for rule in text_rules:
            result = rule(line_number, line, bib_text)
            if not result:
                warnings.append(LintWarning(bibliography, line_number, rule))


def _apply_entry_rules(bibliography, bib_data, bib_text, entry_rules, warnings):
    for key, entry in bib_data.entries.items():
        for rule in entry_rules:
            line_number, offset = find_entry_line_number(bib_text, key)
            result = rule(key, entry, bib_data)
            if not result:
                warnings.append(LintWarning(bibliography, line_number, rule))


@dataclass
class LintWarning:
    file: str
    line: int
    rule: Rule

    def log(self):
        logger = logging.getLogger()
        msg = "{}:{} {}".format(self.file, self.line,str(self.rule))
        logger.warning(msg)



import re

import markdown_table

MONTH_NAMES = {
    'jan': 'jan',
    'feb': 'feb',
    'mar': 'mar',
    'apr': 'apr',
    'may': 'may',
    'jun': 'jun',
    'jul': 'jul',
    'aug': 'aug',
    'sep': 'sep',
    'oct': 'oct',
    'nov': 'nov',
    'dec': 'dec'
}


def find_match_line_number(text, pattern, group):
    regex = re.compile(pattern)
    match = next(regex.finditer(text))
    start = match.start(group)
    lineno = text.count('\n', 0, start)
    if lineno:
        offset = start - text.rfind('\n', 0, start)
    else:
        offset = start
    return lineno + 1, offset + 1


def find_entry_line_number(text, key):
    pattern = r'\s*@[a-zA-Z]+\s*{\s*(' + key + ')\s*,'
    return find_match_line_number(text, pattern, 1)


def format_rules_markdown_tables(rules):
    result = "# bibL rules\n"
    headers = ["Rule ID", "Rule description"]
    matrix = []
    for i, rule in enumerate(rules):
        if i != 0 and rule.id[0] != rules[i - 1].id[0]:
            result += "## " + rules[i - 1].id[0]
            result += "\n"
            result += markdown_table.render(headers, matrix)
            result += "\n\n"
            matrix = []
        matrix.append([f'`{rule.id}`', rule.description])
    result += "## " + rule.id[0]
    result += "\n"
    result += markdown_table.render(headers, matrix)
    return result



import re

from fuzzywuzzy import fuzz
from unidecode import unidecode

from bibl.rule import register_entry_rule, register_text_rule



@register_entry_rule('D00', 'Entry not in alphabetical order by key')
def keys_alphabetical(key, entry, database):

    keys = list(database.entries.keys())
    entry_num = keys.index(key)
    if entry_num == len(keys) -1:
        return True
    else:
        return entry.key.lower() <= keys[entry_num+1].lower()

@register_text_rule('D01', 'Preamble should begin at first line of document')
def line_length(line_number, line, text):
    regex = re.compile(r'^\s*@preamble')
    return not regex.match(line.lower()) or line_number == 0

# TODO: check fuzzy author name consistency


# TODO fuzzy match on entire entry
@register_entry_rule('D02', 'Possible duplicate entry based on similar titles')
def title_duplicate(key, entry, database):
    if not 'title' in entry.fields:
        return True
    for e in database.entries.values():
        if not 'title' in e.fields:
            continue
        t1 = unidecode(entry.fields['title']).lower()
        t2 = unidecode(e.fields['title']).lower()
        if e != entry and fuzz.partial_ratio(t1, t2) > 90:
            return False
    return True



import re

from bibl.config import get_config
from bibl.rule import register_text_rule


@register_text_rule('T00', 'Non-ascii character')
def ascii_chars(line_number, line, text):
    return line.isascii()


@register_text_rule('T01', 'Non-standard whitespace at beginning of line (indents should be {} spaces)'.format(
    get_config()['indent_spaces']))
def trailing_whitespace(line_number, line, text):
    regex = re.compile(r'(^ {' + str(get_config()['indent_spaces']) + r'}\S)|(^\S)|(^$)')
    return bool(regex.match(line))


@register_text_rule('T02', 'Whitespace at end of line')
def ending_whitespace(line_number, line, text):
    regex = re.compile(r'(^.*\S$)|(^$)')
    return bool(regex.match(line))


@register_text_rule('T03', 'Line length exceeds {} characters'.format(get_config()['max_line_length']))
def line_length(line_number, line, text):
    return len(line) <= get_config()['max_line_length']

# TODO: invalid syntax
# TODO: preamble first line
# TODO: string directly after preamble



import itertools
import os
import re
from unidecode import unidecode

from bibl.config import get_config
from bibl.rule import register_entry_rule
from bibl.text_utils import MONTH_NAMES


# TODO: use of undefined symbol



@register_entry_rule('E00', 'Keys of published works should have format AuthorYEARa')
def key_format(key, entry, database):
    if not 'year' in entry.fields:
        return True
    author = entry.persons['author'][0]
    names = author.rich_prelast_names + author.rich_last_names
    correct_key_unicode = "".join([str(name.capitalize()) for name in names]) + entry.fields['year']
    correct_key_ascii = unidecode(correct_key_unicode)
    regex = re.compile(correct_key_ascii + r'[a-zA-Z]?')
    return bool(regex.match(key))


@register_entry_rule('E01', f'Author first names should not be abbreviated')
def author_first_name_abbr(key, entry, database):
    for person in itertools.chain(*entry.persons.values()):
        for name in person.first_names:
            if len(name) == 1:
                return False
    return True


if get_config()['abbreviation_dot']:
    author_middle_name_abbr_desc = "Author middle names should be abbreviated with ."
else:
    author_middle_name_abbr_desc = "Author middle names should be abbreviated without ."


@register_entry_rule('E02', author_middle_name_abbr_desc)
def author_middle_name_abbr(key, entry, database):
    for person in itertools.chain(*entry.persons.values()):
        for name in person.middle_names:
            if get_config()['abbreviation_dot'] and len(name) == 1:
                return False
            if not get_config()['abbreviation_dot'] and name.endswith('.'):
                return False
    return True


@register_entry_rule('E03', 'The usage of `et al.` in the author field should be replaced by a list of all authors')
def author_et_al(key, entry, database):
    if 'author' in entry.fields:
        return not entry.fields['author'].lower().contains('et al')
    else:
        return True


def _process_file_path(file):
    if ':' in file:
        return file.split(':')[1]
    else:
        return file


@register_entry_rule('E04', 'Files should be linked with relative path')
def file_relative_path(key, entry, database):
    if 'file' in entry.fields:
        path = _process_file_path(entry.fields['file'])
        return not os.path.isabs(path)
    else:
        return True


@register_entry_rule('E05', 'Linked file is not present')
def file_present(key, entry, database):
    if 'file' in entry.fields:
        path = _process_file_path(entry.fields['file'])
        if os.path.isabs(path):
            return os.path.exists(path)
        else:
            abs_path = os.path.join(os.path.dirname(database.file), path)
            return os.path.exists(abs_path)
    else:
        return True


@register_entry_rule('E06', 'Incorrect doi format')
def doi_format(key, entry, database):
    if 'doi' in entry.fields:
        regex = re.compile(r'^10.\d{4,9}/[-._;()/:a-z0-9]+$')
        return bool(regex.match(entry.fields['doi']))
    else:
        return True


@register_entry_rule('E07', 'Incorrect ISBN format')
def isbn_format(key, entry, database):
    if 'isbn' in entry.fields:
        regex = re.compile(r'^(?=(?:\D*\d){10}(?:(?:\D*\d){3})?$)[\d-]+$')
        return bool(regex.match(entry.fields['isbn']))
    else:
        return True


@register_entry_rule('E08', 'Page numbers should have format xx--yy')
def page_format(key, entry, database):
    if 'page' in entry.fields:
        regex = re.compile(r'^\d+(--\d+)?$')
        return bool(regex.match(entry.fields['page']))
    else:
        return True


@register_entry_rule('E09', 'End page larger than start page. Page numbers in aaaa--bb should be written as aaaa--aabb')
def page_format_ascending(key, entry, database):
    if 'page' in entry.fields:
        regex = re.compile(r'^\d+--\d+$')
        if not regex.match(entry.fields['page']):
            return True
        else:
            groups = entry.fields['page'].split('--')
            return int(groups[1]) >= int(groups[0])
    else:
        return True


@register_entry_rule('E10', 'Month should be in 3-letter lowercase format')
def month_format(key, entry, database):
    if 'month' in entry.fields:
        return entry.fields['month'] in MONTH_NAMES.keys()
    else:
        return True



import itertools

from bibl.config import get_config
from bibl.rule import register_entry_rule


@register_entry_rule('M00', 'No authors or editors found')
def authors(key, entry, database):
    return bool(list(itertools.chain(*entry.persons.values())))


for entry_type, spec in get_config()['type_spec'].items():
    for req_field in spec['required']:
        id = 'M01{}{}'.format(entry_type.capitalize(), req_field.capitalize())


        @register_entry_rule(id, 'Missing required field `{}` for entry type `{}`'.format(req_field, entry_type))
        def check_required_field_present(key, entry, database, entry_type=entry_type, req_field=req_field):
            if entry.type == entry_type:
                return req_field in entry.fields
            else:
                return True

    for opt_field in spec['optional']:
        id = 'M02{}{}'.format(entry_type.capitalize(), opt_field.capitalize())


        @register_entry_rule(id, 'Missing optional field `{}` for entry type `{}`'.format(opt_field, entry_type))
        def check_optional_field_present(key, entry, database, entry_type=entry_type, opt_field=opt_field):
            if entry.type == entry_type:
                return opt_field in entry.fields
            else:
                return True



import os
import warnings

import click
from bibl import __version__
from bibl.lint import lint as bibl_lint
from bibl.config import load_config, set_config
from bibl.rule import load_rules
from bibl.text_utils import format_rules_markdown_tables


@click.group()
@click.option('-c', '--config', help='Custom configuration file path.', type=str)
@click.option('--select', help='Comma separated list of enabled rules, all other rules will be disabled.', type=str)
@click.option('--ignore', help='Comma separated list of disabled rules, all other rules will be enabled.', type=str)
@click.option('--indent-spaces', help='Number of trailing whitespaces for indented line, used by TO1.', type=int)
@click.option('--max-line-length', help='Max line length before wrap recommended, used by T03.', type=int)
@click.option('--abbreviation-dot', help='Abbreviate middle names with dot.', is_flag=True)
def cli(config, select, ignore, indent_spaces, max_line_length, abbreviation_dot):
    if not config is None:
        load_config(config)
    elif os.path.isfile('.bibl.yml'):
        load_config('.bibl.yml')

    if not select is None:
        set_config('select', select.split(','))
    if not ignore is None:
        set_config('ignore', ignore.split(','))
    set_config('indent_spaces', indent_spaces)
    set_config('max_line_length', max_line_length)
    set_config('abbreviation_dot', abbreviation_dot)


@cli.command(help="Lint a BibTeX bibliography file.")
@click.argument('bibliography', type=str, nargs=-1)
def lint(bibliography):
    warnings.filterwarnings("ignore")
    for bib in bibliography:
        bibl_lint(bib)


@cli.command(help="Show all available rules.")
@click.option('-m', 'markdown', help='Format rules as markdown table.', is_flag=True)
def list_all(markdown):
    rules = load_rules().all
    if markdown:
        click.echo(format_rules_markdown_tables(rules))
    else:
        for rule in rules:
            click.echo(rule)


@cli.command(help="Show all rules enabled by the configuration.")
@click.option('-m', 'markdown', help='Format rules as markdown table.', is_flag=True)
def list_enabled(markdown):
    rules = load_rules().enabled
    if markdown:
        click.echo(format_rules_markdown_tables(rules))
    else:
        for rule in rules:
            click.echo(rule)

@cli.command(help="Show the package version.")
def version():
    click.echo('bibL version: ' + __version__)

if __name__ == '__main__':
    cli(prog_name='bibl')



import fnmatch
import os
from typing import Callable, Dict, List
from importlib import import_module
from pybtex.database import Entry

from bibl.config import get_config


class Rule():

    def __init__(self, id: str, description: str):
        self.id = id
        self.description = description
        self._rule = None

    def __str__(self):
        return "{}: {}".format(self.id, self.description)

    def __call__(self, *args, **kwargs):
        return self._rule(*args, **kwargs)

    @property
    def enabled(self):
        if get_config()['select']:
            for pattern in get_config()['select']:
                if fnmatch.fnmatch(self.id, pattern):
                    return True
            return False
        if get_config()['ignore']:
            for pattern in get_config()['ignore']:
                if fnmatch.fnmatch(self.id, pattern):
                    return False
            return True
        return True


class EntryRule(Rule):

    def __init__(self, id, description, rule: Callable[[str, Entry, Dict[str, Entry]], bool], **kwargs):
        super().__init__(id, description)
        self._rule = rule


class TextRule(Rule):

    def __init__(self, id, description, rule: Callable[[int, str, str], bool], **kwargs):
        super().__init__(id, description)
        self._rule = rule


class RuleStore:

    def __init__(self):
        self._rules = []

    def register(self, rule: Rule):
        position = 0
        while position < len(self._rules) and self._rules[position].id < rule.id:
            position += 1
        self._rules.insert(position, rule)

    @property
    def all(self) -> List[Rule]:
        return self._rules

    @property
    def enabled(self) -> List[Rule]:
        return [rule for rule in self._rules if rule.enabled]

    @property
    def enabled_entry_rules(self) -> List[EntryRule]:
        return [rule for rule in self.enabled if isinstance(rule, EntryRule)]

    @property
    def enabled_text_rules(self) -> List[TextRule]:
        return [rule for rule in self.enabled if isinstance(rule, TextRule)]


_ALL_RULES: RuleStore = RuleStore()


def register_entry_rule(id, description: str):
    def decorator(f: Callable[[str, Entry, Dict[str, Entry]], bool]):
        rule = EntryRule(id, description, f)
        _ALL_RULES.register(rule)

    return decorator


def register_text_rule(id: str, description: str):
    def decorator(f: Callable[[int, str, str], bool]):
        rule = TextRule(id, description, f)
        _ALL_RULES.register(rule)

    return decorator


def load_rules() -> RuleStore:
    for module in os.listdir(os.path.join(os.path.dirname(__file__), 'rules')):
        if module == '__init__.py' or module[-3:] != '.py':
            continue
        __import__('bibl.rules.' + module[:-3], locals(), globals())
    del module
    return _ALL_RULES



import unittest

from bibl.lint import lint


class TestBase(unittest.TestCase):

    def test_run(self):
        lint("test_data/mit.bib", output=False)

    def test_jabref(self):
        lint("test_data/bci.bib", output=False)

    def test_incorrect_syntax(self):
        lint("test_data/syntax.bib", output=False)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup

with open("README.md", 'r') as readme:
    long_description = readme.read()

from bibl import __version__

setup(
    name='bibL',
    version=__version__,
    description='A minimalistic bibtex linter',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/arne.vandenkerchove/biblint',
    author='Arne Van Den Kerchove',
    author_email='arne@vandenkerchove.com',
    packages=['bibl', 'bibl.rules'],
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'click>=7,<8',
        'fuzzywuzzy>=0.10,<1',
        'markdown-table',
        'pybtex==0.23.0',
        'pyyaml>=5,<6',
        'unidecode>=1,<2',
    ],
    extras_require={
        'dev': [
            'flake8',
            'anybadge',
            'markdown',
            'check-manifest',
            'twine',
        ]
    },
    entry_points={
        "console_scripts": ["bibl=bibl.__main__:cli"],
    },
)



from bibl.config import get_config
from bibl.rule import register_entry_rule


@register_entry_rule('U00', 'Unrecognized entry type')
def entry_type(key, entry, database):
    return entry.type in get_config()['type_spec'].keys()


for entry_type, spec in get_config()['type_spec'].items():
    id = 'U01{}'.format(entry_type.capitalize())

    @register_entry_rule(id, 'Unrecognized field type for entry type `{}`'.format(entry_type))
    def field_type(key, entry, database, entry_type=entry_type):
        if entry.type == entry_type and entry.type in get_config()['type_spec'].keys():
            optional_fields = get_config()['type_spec'][entry.type]['optional']
            required_fields = get_config()['type_spec'][entry.type]['required']
            for field_type in entry.fields.keys():
                if not field_type in optional_fields and not field_type in required_fields:
                    return False
        return True
